from botiverse.bots.BasicBot.BasicBot import BasicBot
from botiverse.bots.WhizBot.WhizBot import WhizBot
from botiverse.bots.ConverseBot.ConverseBot import ConverseBot
from botiverse.bots.BasicTaskBot.BasicTaskBot import BasicTaskBot
from botiverse.bots.TaskBot.TaskBot import TaskBot
from botiverse.bots.VoiceBot.SpeechClassifier import SpeechClassifier
from botiverse.bots.VoiceBot.VoiceBot import VoiceBot
